/* Bibliothek MyCAN
 *  
 * Zur Kommunikation mehrerer Arduinos über eine Eindrahtleitung
 * 
 * Unterstützt angelehnt an den CAN-BUS
 * - Prioritätssteuerung der Nachrichten nach ID
 * - Multi-Master-System: alle können gleichzeitig senden
 * - Fehlerkontrolle per Parity Bits
 *  
 *  Message Structure
    ID Value Parity
    10   16     4 Bits = 30 Bits
   3FF  FFFF
*/


// Debug
#define MCDEBUG

// defines
#define ULONG unsigned long

// define MyCAN timings in ms
#define MC_TIMEBASE_1  4 // min: 8
#define MC_TIMEBASE_0  8
#define MC_TIMEBASE_B  2 // Break

// MyCAN ports
const int MC_Pin_Send       =  11;
const int MC_Pin_Read       =  12;
const int MC_Pin_ErrLED     =  13;
const int MC_Pin_StatusLED  =  13;

// global variables
byte  bitstreamOut[30];
ULONG MC_RXqueue[32];
byte  MC_RXmessagesInQueue = 0;
byte  MC_RXmessagesInQueueMax = 0;
ULONG MC_lastBusActivity = 0;
// statistics
ULONG MCcountSameState = 0;
ULONG MCcountMillisOverflow = 0;
ULONG MCcountOnes = 0;
ULONG MCcountZeros = 0;
ULONG MCcountTimingError = 0;
ULONG MCcountMessagesRXInterrupted = 0;
ULONG MCcountMessagesRXParityFail = 0;
ULONG MCcountMessagesRXok = 0;
ULONG MCcountMessagesRXtime = 0;
ULONG MCcountMessagesTX = 0;
ULONG MCcountMessagesTXoverwritten = 0;
ULONG MCbusWaitTime = 0;

// Setup for MyCAN
void MC_setup()
{
  pinMode(MC_Pin_Send, OUTPUT);
  pinMode(MC_Pin_ErrLED, OUTPUT);
  pinMode(MC_Pin_Read, INPUT);
  randomSeed(analogRead(0));
  // turn on pin change interrupt on port B
  cli();
  PCICR = 0b00000001;
  PCMSK0 = 0b0010000; // turn on pin PB4 = D12
  sei();
  delay (3000);   // wait 3s to start Hterm
  #ifdef MCDEBUG
    Serial.begin(9600);
    while (!Serial) { }
    // prints title with ending line break
    Serial.println("--- Restarting ---");
  #endif
}

// MyCAN receiver Interrupt Service Routine
ISR(PCINT0_vect)
{
  static int lastBusValue = 1;
  static ULONG startBitTime;
  static ULONG startMessageTime;
  static byte bitPosition = 0;
  static byte RXmessage[30];
  ULONG deltaT;

  int busValue = digitalRead(MC_Pin_Read);

  // handle same state as last time
  if (busValue == lastBusValue)
  {
    MCcountSameState++;
    return;
  }
  // handle state changes
  if (busValue == 0) // transition from HI=>LO: start of bit
  {
    startBitTime = millis();
  }
  else // transition from LO=>HI: bit finished
  {
    // calculate deltaT including overflow
    if (millis() >= startBitTime) 
      deltaT = millis()-startBitTime;
    else // millis() overflow (every ~49 days)
    {
      deltaT = 0xFFFFFFFF-startBitTime;
      deltaT += millis();
      MCcountMillisOverflow++;
    }
    // analyze received bit
    if (deltaT >= MC_TIMEBASE_B*9) // start of message
    {
      if (bitPosition > 0) // interrupting previous message
      {
        MCcountMessagesRXInterrupted++;
        bitPosition = 0;
      }
      startMessageTime = millis()-MC_TIMEBASE_B*10;
    }
    else if ((deltaT >= MC_TIMEBASE_1 * 3/4) && (deltaT <= MC_TIMEBASE_1 * 5/4)) // received a "1"
    {
      RXmessage[bitPosition] = 1;
      bitPosition++;
      MCcountOnes++;
    }
    else if ((deltaT >= MC_TIMEBASE_0 * 3/4) && (deltaT <= MC_TIMEBASE_0 * 5/4)) // received a "0"
    {
      RXmessage[bitPosition] = 0;
      bitPosition++;
      MCcountZeros++;  
    }
    else // timing error, reset RX message
    {
      bitPosition = 0; 
      MCcountTimingError++;
    }
  }
  // analyse finished message
  if (bitPosition == 30)
  {
    if (MC_checkParity(RXmessage))
    {
      MC_addMessageToRXqueue(RXmessage);
      MCcountMessagesRXtime += millis()-startMessageTime;
      MCcountMessagesRXok++;
    }
    else
      MCcountMessagesRXParityFail++;
      
    // reset message 
    bitPosition = 0;
  }
  lastBusValue = busValue;
  MC_lastBusActivity = millis();
}

void MC_blink_LED(int LED, int msHigh, int msLow)
{
    digitalWrite(LED, HIGH);
    delay(msHigh);
    digitalWrite(LED, LOW);
    delay(msLow);
}

void MC_DebugLine(char* text, long value, int type)
{
  #ifdef MCDEBUG
    Serial.print(millis());
    Serial.print(": ");
    Serial.print(text);
    if (value != -1) 
    {  
      Serial.print(" ");    
      Serial.print(value, type);
    }
    Serial.println("");
  #endif
}

void MC_fatal_error(int errorID)
{
  while (1) // never stop blinking
  {
    // first Blink long
    MC_blink_LED(MC_Pin_ErrLED, 1000, 200);
    // Blink according to errorID
    for (int i = 0; i < errorID; i++)
    {
      MC_DebugLine("FATAL ERROR ID", errorID, DEC); 
      for (int t = 0; t < 20; t++)      
        MC_blink_LED(MC_Pin_ErrLED, 300, 200);
    }
  }
}

ULONG MC_calcParity(ULONG message)
{
  ULONG dValue = 0x20000000, outMessage = 0;
  int   bitnum = 0, i, p, sum;
  //MC_DebugLine("message",long(message),HEX);
  //MC_DebugLine("message",long(message),BIN);
  // calc bitstreamOut
  while (dValue > 0)
  {
    if (message >= dValue)
    {
      message -= dValue;
      bitstreamOut[bitnum] = 1;
    }
    else bitstreamOut[bitnum] = 0;
    dValue /= 2;
    bitnum++;
  }
  if (bitnum > 30)
    MC_fatal_error(3);
  // calc 4 prities
  for (p = 0; p < 4; p++)
  {
    sum = 0;
    for (i = p; i < 26; i+= 4)
      sum += bitstreamOut[i];
    bitstreamOut[p+26] = sum%2;
  }
  // recalc unsigned long
  for (i = 0; i < 30; i++)
    outMessage += long(bitstreamOut[i]) << (29-i);
  //MC_DebugLine("message",long(outMessage),HEX);
  return outMessage;
}

byte MC_checkParity(byte RXmessage[30])
{
  int   i, p, sum;
  // calc 4 prities
  for (p = 0; p < 4; p++)
  {
    sum = 0;
    for (i = p; i < 26; i+= 4)
      sum += RXmessage[i];
    sum += RXmessage[26+p];
    if (sum%2 > 0)
      return false;
  }
  return true;
}


void MC_sendMessage(ULONG ID, ULONG Value)
{
  /* Message Structure
      ID Value Parity
      10   16     4 Bits = 30 Bits
     3FF  FFFF
  */
  ULONG message, pMessage, startWaiting;
  int   i, messageSent = false;
  // check size of parameters
  if (ID    >= 1024)  MC_fatal_error(1);
  if (Value >= 65536) MC_fatal_error(2);
  // concat message
  message = ID << 20;
  message += Value << 4;
  pMessage = MC_calcParity(message); // bitstreamOut holds Message now
  #ifdef MCDEBUG
    MC_DebugLine("sending message:  ",long(pMessage),HEX);
  #endif
  // wait after last message
  int busAvailable = false;
  startWaiting = millis();
  while (not busAvailable)
  {
    while (millis()-MC_lastBusActivity < MC_TIMEBASE_B*10+MC_TIMEBASE_0+MC_TIMEBASE_1) 
      delayMicroseconds(random(100,3000));
    // wait until bus is available
    while (digitalRead(MC_Pin_Read) == 0)
      delayMicroseconds(random(50,2000));
    if (millis()-MC_lastBusActivity >= MC_TIMEBASE_B*10+MC_TIMEBASE_0+MC_TIMEBASE_1)
      busAvailable = true;  
  }
  MCbusWaitTime += millis()-startWaiting;
  // repeat sending until successfully done
  do 
  {
    // send long low to declare start of message
    digitalWrite(MC_Pin_Send, 1); 
    delay(MC_TIMEBASE_B*10); // total MC_TIMEBASE_B*10
    digitalWrite(MC_Pin_Send, 0); 
    delay(MC_TIMEBASE_B);
    // check if bus overwritten dominantly
    if (digitalRead(MC_Pin_Read) == 0)
    {
      MCcountMessagesTXoverwritten++;
      continue;
    }
    // send message itself (30 Bits)
    for (i=0; i < 30; i++)
    {
      // set output pin high => bus low
      digitalWrite(MC_Pin_Send, 1); 
      // wait (information in low-time)
      if (bitstreamOut[i] == 0)
        delay(MC_TIMEBASE_0);
      else
        delay(MC_TIMEBASE_1);
      // put bus pack high
      digitalWrite(MC_Pin_Send, 0); 
      delay(MC_TIMEBASE_B);
      // check if bus overwritten dominantly
      if (digitalRead(MC_Pin_Read) == 0)
      {
        MCcountMessagesTXoverwritten++;
        continue;
      }
    }
    messageSent = true;
  }
  while (messageSent == false);
  MCcountMessagesTX++;
}

String MC_strLen(String strIn, int len)
{
  String strOut = strIn;
  while (strOut.length() < len)
    strOut = " "+strOut;
  return strOut; 
}

String MC_strLen(ULONG number, int len)
{
  String strOut = String(number);
  while (strOut.length() < len)
    strOut = " "+strOut;
  return strOut; 
}

void MC_printStats()
{
  float perc, x;
  String line;
  
  Serial.println("------------ MyCAN Statistics -------------");
  Serial.print("MCcountSameState:         ");
  Serial.println(MC_strLen(MCcountSameState,10));

  Serial.print("MCcountMillisOverflow:    ");
  Serial.println(MC_strLen(MCcountMillisOverflow, 10));

  Serial.print("MCcountOnes:              ");
  Serial.println(MC_strLen(MCcountOnes, 10));

  Serial.print("MCcountZeros:             ");
  Serial.println(MC_strLen(MCcountZeros, 10));

  Serial.print("MCcountTimingError:       ");
  Serial.println(MC_strLen(MCcountTimingError, 10));

  perc = 100.0*MCcountMessagesRXInterrupted/(MCcountMessagesRXok+MCcountMessagesRXInterrupted+MCcountMessagesRXParityFail);
  line = "MCcountMessagesRXInterrupted: "+MC_strLen(MCcountMessagesRXInterrupted, 6)+ " "+MC_strLen(String(perc,1),5)+"%";
  Serial.println(line);

  perc = 100.0*MCcountMessagesRXParityFail/(MCcountMessagesRXok+MCcountMessagesRXInterrupted+MCcountMessagesRXParityFail);
  line = "MCcountMessagesRXParityFail:  "+MC_strLen(MCcountMessagesRXParityFail, 6)+ " "+MC_strLen(String(perc,1),5)+"%";
  Serial.println(line);

  perc = 100.0*MCcountMessagesTX/(MCcountMessagesRXok+MCcountMessagesRXInterrupted+MCcountMessagesRXParityFail+MCcountMessagesTX);
  line = "MCcountMessagesTX:            "+MC_strLen(MCcountMessagesTX, 6)+ " "+MC_strLen(String(perc,1),5)+"%";
  Serial.println(line);


  perc = 100.0*MCcountMessagesRXok/(MCcountMessagesRXok+MCcountMessagesRXInterrupted+MCcountMessagesRXParityFail);
  line = "MCcountMessagesRXok:          "+MC_strLen(MCcountMessagesRXok, 6)+ " "+MC_strLen(String(perc,1),5)+"%";
  Serial.println(line);

  x = 1.0*MCcountMessagesRXtime/MCcountMessagesRXok;
  line = "MCcountMessagesRXtime:        "+MC_strLen(String(x,1),6)+" ms";
  Serial.println(line);

  line = "MC_RXmessagesInQueueMax:      "+MC_strLen(MC_RXmessagesInQueueMax, 6);
  Serial.println(line);

  line = "MCcountMessagesTXoverwritten: "+MC_strLen(MCcountMessagesTXoverwritten, 6);
  Serial.println(line);

  x = 1.0*MCbusWaitTime/1000;
  perc = 100.0*MCbusWaitTime/millis();
  line = "MCbusWaitTime:                "+MC_strLen(String(x,1),6)+" s "+String(perc,1)+"%";
  Serial.println(line);

  x = 1.0*MCbusWaitTime/MCcountMessagesTX;
  line = "MCbusWaitTime per Message:    "+MC_strLen(String(x,1),6)+" ms";
  Serial.println(line);

  x = 1000.0*MCcountMessagesRXok/millis();
  line = "Messages per Second:          "+MC_strLen(String(x,2),6);
  Serial.println(line);

  x = 3600000.0*MCcountMessagesRXok/millis();
  line = "Messages per hour:            "+MC_strLen(String(x,0),6);
  Serial.println(line);


  perc = 100.0*MCcountMessagesRXtime/millis();
  line = "Bus Load:                     "+MC_strLen(String(perc,1),6)+"%";
  Serial.println(line);

  Serial.println("-------------------------------------------");
}

void MC_addMessageToRXqueue(byte RXmessage[30])
{
  int i;
  ULONG decValue = 0;
  for (i = 0; i < 26; i++)
  {
    decValue += long(RXmessage[i]) << (25-i);
  }
  MC_DebugLine("message received: ",long(decValue),HEX);
  MC_RXqueue[MC_RXmessagesInQueue] = decValue;
  // avoid overflow
  if (MC_RXmessagesInQueue < 31)
    MC_RXmessagesInQueue++;  
  // determine all time maximum of messages in queue
  MC_RXmessagesInQueueMax = max(MC_RXmessagesInQueueMax, MC_RXmessagesInQueue);
}

ULONG MC_getMessageFromRXqueue()
{
  ULONG message = 0xFFFFFFFF;
  int i;
  if (MC_RXmessagesInQueue > 0)
  {
    message = MC_RXqueue[0];
    // shift queue (FIFO)
    for (i = 0; i < 15; i++)
      MC_RXqueue[i] = MC_RXqueue[i+1];
    MC_RXmessagesInQueue--;
   }
   return message; 
}
